import argparse
import os

from dotenv import load_dotenv, dotenv_values

from banyan_extract import NemoparseProcessor

try:
    from banyan_extract import MarkerProcessor
except:
    pass

def parse_arguments():
    parser = argparse.ArgumentParser()
    parser.add_argument("input_file", default=None, type=str, help="Path for a single file to be processed")
    parser.add_argument("output_dir", default=None, type=str, help="Path for output from single or multiple files")
    parser.add_argument("--is_input_dir", action="store_true",  help="Flags to set input file to directory")
    parser.add_argument("--output_base", default="banyan-extract-output", type=str,  help="Base name for output files")
    parser.add_argument("--backend", default="nemoparse", type=str,  help="Which backend to use (nemoparse (default)- Nemotron Parse, marker - marker)")
    parser.add_argument("--config_file", default=".env", type=str,  help="Which config file to use (defaults to ./.env)")
    parser.add_argument("--endpoint", default="", type=str, help="Endpoint url for nemoretreiver-parse model")
    parser.add_argument("--model_name", default="", type=str, help="Endpoint url for nemoretreiver-parse model")
    parser.add_argument("--checkpointing", action="store_true", help="Flag where if true, then batch documents will be saved as they get processed")
    parser.add_argument("--draw_bboxes", action="store_true", default=False, help="Flag where if true, then ouptut will include images that show most bboxes found")
    return parser.parse_args()


def main():
    args = parse_arguments()

    output_directory = args.output_dir
    output_base = args.output_base
    endpoint = args.endpoint
    model_name = args.model_name
    backend = args.backend

    if args.backend == "nemoparse":
        if len(endpoint) == 0:
            config_values = dotenv_values(args.config_file)
            endpoint = config_values["NEMOPARSE_ENDPOINT"]
            model_name = config_values["NEMOPARSE_MODEL"]
            print(f"Using endpoint: {endpoint}")
            print(f"Using model: {model_name}")

        if endpoint != "":
            document_processor = NemoparseProcessor(endpoint_url=endpoint, model_name=model_name)
        else:
            raise Exception("Missing nemotron-parse endpoint url!")
    elif args.backend == "marker":
        document_processor = MarkerProcessor()

    if args.is_input_dir:
        input_directory = args.input_file

        file_paths = []
        basenames = []
        for root, _, files in os.walk(input_directory):
            for filename in files:
                file_paths.append(os.path.join(root, filename))
                basenames.append(os.path.basename(filename))

        outputs = document_processor.process_batch_documents(file_paths, use_checkpointing=args.checkpointing, draw_bboxes=args.draw_bboxes, output_dir=output_directory)
        if not args.checkpointing:
            for file_output, basename in zip(outputs, basenames):
                file_output.save_output(output_directory, basename)
    else:
        filename = args.input_file
        outputs = document_processor.process_document(filename)

        outputs.save_output(output_directory, output_base)


if __name__ == '__main__':
    main()



from papermage.magelib import Document
from papermage.recipes import CoreRecipe
from papermage.visualizers.visualizer import plot_entities_on_page
import pathlib
import numpy as np

from .processor import Processor
from ..output.papermage_output import PaperMageOutput

class PaperMageProcessor(Processor):
    
    def __init__(self):
        super().__init__()
        self.recipe = CoreRecipe()

    def process_document(self, mode, filepath, options=None, colors=None):
        pdf_path = pathlib.Path(filepath)

        if mode == 'bound_single':
            output = []

            doc = self.recipe.from_pdf(pdf=pdf_path)

            for page_num, page in enumerate(doc.pages):
                intersect_dict = {}

                for option in options:
                    intersect_dict[option] = page.intersect_by_box(option)

                plotted = plot_entities_on_page(page_image=doc.images[page_num], entities=intersect_dict[options[0]], box_color=colors[0])

                for idx, option in enumerate(options[1:]):
                    plotted = plot_entities_on_page(page_image=plotted, entities=intersect_dict[option], box_color=colors[idx])
                    
                output.append(plotted)

        elif mode == 'extract_single':
            output = self.recipe.run(pdf_path)
            
        return PaperMageOutput(output)


    def process_batch_documents(self, mode, filepaths, options=None, colors=None):
        parent_path = pathlib.Path(filepaths)
        file_list = list(pathlib.Path(parent_path).glob('*.pdf'))

        if mode == 'bound_batch':
            output = {}

            for filename in file_list:
                filename_strip = str(filename.name).replace('.pdf', '')
                output[filename_strip] = []

                doc = self.recipe.from_pdf(pdf=filename)

                for page_num, page in enumerate(doc.pages):
                    intersect_dict = {}

                    for option in options:
                        intersect_dict[option] = page.intersect_by_box(option)

                    plotted = plot_entities_on_page(page_image=doc.images[page_num], entities=intersect_dict[options[0]], box_color=colors[0])

                    for idx, option in enumerate(options[1:]):
                        plotted = plot_entities_on_page(page_image=plotted, entities=intersect_dict[option], box_color=colors[idx])
                    
                    output[filename_strip].append(plotted)

        elif mode == 'extract_batch':
            output = {}

            for filename in file_list:
                filename_strip = str(filename.name).replace('.pdf', '')
                output[filename_strip]=self.recipe.run(filename)

        return PaperMageOutput(output)
    


import csv 
import os
import json

from PIL import Image, ImageDraw
from dataclasses import dataclass

from .output import ModelOutput
from ..converter import convert_latex_table_to_csv


@dataclass
class NemoparseData:
    text: list 
    bbox_json: str
    images: list
    tables: list
    bbox_image: Image


class NemoparseOutput(ModelOutput):

    def __init__(self):
        super().__init__()
        self.text: list[str] = []
        self.images: list[list] = []
        self.tables: list[list] = []
        self.bboxdata: list[str] = []
        self.bbox_image: list[Image] = []
        
    def add_output(self, output_data):
        self.text.append(output_data.text)
        self.images.append(output_data.images)
        self.tables.append(output_data.tables)
        self.bboxdata.append(output_data.bbox_json)
        self.bbox_image.append(output_data.bbox_image)

    def save_output(self, output_directory, filename_base):
        img_index = 0
        img_filenames = []
        for image_list in self.images:
            for img in image_list:
                img_filename = f"{filename_base}_image_{img_index}.png"
                try:
                    img.save(os.path.join(output_directory, img_filename))
                except Exception as e:
                    print(f"An error occurred trying to save the image: {img_filename}: {e}")

                img_index += 1
                img_filenames.append(img_filename)

        with open(
            os.path.join(output_directory, f"{filename_base}.md"), "w+") as f:
            img_index = 0
            for text_list in self.text:
                for text in text_list:
                    if "![{}]({})" in text:
                        text = text.format(f"Image {img_index}", img_filenames[img_index])
                        img_index += 1
                    f.write(text + "\n\n")
                f.write("\n")

        with open(
            os.path.join(output_directory, f"{filename_base}_bbox.json"),
            "w+") as f:
            for bboxdata in self.bboxdata:
                f.write(json.dumps(bboxdata, indent=2))

        for index, bbox_image in enumerate(self.bbox_image):
            bbox_image.save(os.path.join(output_directory, f"{filename_base}_bbox_image_{index}.png"))

        table_index = 0
        for table_list in self.tables:
            for table in table_list:
                table_name = f"{filename_base}_table_{table_index}.csv"
                converted_table = convert_latex_table_to_csv(table)

                with open(os.path.join(output_directory, table_name), 'w') as csv_file:
                    csv_writer = csv.writer(csv_file)
                    for row in converted_table:
                        csv_writer.writerow(row)

                table_index += 1

    def get_bbox_output(self, with_bbox_data=True):
        dict_data = {}
        for i, data in enumerate(self.bboxdata):
            if with_bbox_data:
                dict_data[f"page_{i}"] = data
            else:
                tmp_data = []
                for entry in data:
                    tmp_entry = {}
                    for key in entry:
                        if key != "bbox":
                            tmp_entry[key] = entry[key]
                    tmp_data.append(tmp_entry)
                dict_data[f"page_{i}"] = tmp_data
        return dict_data
        #return self.bboxdata

    def get_output_as_json(self, with_bbox_data=True):
        if with_bbox_data:
            return json.dumps(self.bboxdata)
        else:
            cleaned_data = []
            for bboxdata in self.bboxdata:
                tmp_data = []
                for entry in data:
                    tmp_entry = {}
                    for key in entry:
                        if key != "bbox":
                            tmp_entry[key] = entry[key]
                    tmp_data.append(tmp_entry)
                cleaned_data.append(tmp_data)
            return json.dumps(cleaned_data)

    def get_output_as_markdown(self):
        full_text = ""
        for page_number, page_text in enumerate(self.text):
            for text in page_text:
                full_text += text
            full_text += "\n"
        return full_text

    def get_content_list(self):
        content = []
        for page_text in self.text:
            tmp_data = ""
            for text in page_text:
                tmp_data += text
            content.append(tmp_data)
        return content

    def get_images(self):
        images = []
        for image_list in self.images:
            for img in image_list:
                images.append(img)
        
        return images




import os
import json

from .output import ModelOutput


class PptxOutput(ModelOutput):
    def __init__(self, text: list, images: list, metadata):
        super().__init__()
        self.text: list = text
        self.images: list = images
        self.metadata = metadata

    def save_output(self, output_directory, filename_base):
        with open(os.path.join(output_directory, f"{filename_base}.md"), "w+") as f:
            for slide_id, slide in enumerate(self.text):
                f.write(f"# Slide {slide_id}\n")
                f.write(slide)
                f.write("\n")

        with open(os.path.join(output_directory, f"{filename_base}_meta.json"), "w+") as f:
            f.write(json.dumps(self.metadata, indent=2))

        for slide_number, slide_images in enumerate(self.images):
            for image_id, image in enumerate(slide_images):
                img_name = f"Slide_{slide_number}_image_{image_id}.png"
                image.save(os.path.join(output_directory, img_name))



import sys
import argparse
import json
import base64
import io
import os

from openai import OpenAI
from PIL import Image, ImageDraw

from .processor import Processor
from ..converter.pdf_to_image import convert_pdf_to_images, convert_bytes_to_images
from ..output.nemoparse_output import NemoparseData, NemoparseOutput

class NemoparseProcessor(Processor):
    
    def __init__(self, endpoint_url="", model_name="nvidia/nemoretriever-parse"):
        super().__init__()
        self.model_url = endpoint_url
        self.client = OpenAI(
              base_url = self.model_url,
              # For local deployment, an API key is not needed but the following string
              # varible should be non-empty.
              api_key = "non-empty"
            )
        self.model=model_name

    def _encode_image(self, image):
        try:
            base64_encoded_data = base64.b64encode(image)
            base64_string = base64_encoded_data.decode("utf-8")
            return base64_string
        except Exception as e:
            print(f"An error occurred trying to encode the document: {e}")
            return None

    def _process_image(self, image, draw_bboxes=True):
        base64_string = self._encode_image(image)
        base64_image = f"data:image/png;base64,{base64_string}"
        messages = [
                {
                    "role": "user",
                    "content": [
                        {
                            "type": "image_url",
                            "image_url": {
                                # You can provide the image URL or send the image
                                # data as a Base64-encoded string.
                                "url": base64_image,
                            },
                        },
                    ]
                }
            ]


        completion = self.client.chat.completions.create(
            model=self.model,
            # See Tool types section for more information.
            tools=[{"type": "function", "function": {"name": "markdown_bbox"}}],
            messages=messages
        )

        tool_call = completion.choices[0].message.tool_calls[0]
        response = json.loads(tool_call.function.arguments)

        bbox_data = response[0]

        base_image = Image.open(io.BytesIO(image))
        width, height = base_image.size

        if draw_bboxes:
            bbox_draw = ImageDraw.Draw(base_image)
            color_dict = {
                        "Text": "red",
                        "Formula": "green",
                        "Code": "blue",
                        "Picture": "magenta",
                        "Table": "cyan",
                        "Caption": "yellow",
                        }

        txt = []
        images = []
        tables = []

        for entry in bbox_data:
            bbox = entry['bbox']
            xmin = bbox['xmin'] * width
            ymin = bbox['ymin'] * height
            xmax = bbox['xmax'] * width
            ymax = bbox['ymax'] * height

            element_text = entry['text']
            if entry['type'] in "Picture":
                cropped_image = base_image.crop((xmin, ymin, xmax, ymax))
                images.append(cropped_image)
                element_text = "![{}]({})"
            elif entry ['type'] in "Table":
                tables.append(entry['text'])
            elif entry['type'] in "Section-header" and "#" not in entry['text']:
                element_text = f"# {entry['text']}"

            txt.append(element_text)

            if draw_bboxes:
                if entry['type'] in color_dict:
                    color = color_dict[entry['type']]
                    if ymin > ymax:
                        ymin, ymax = ymax, ymin
                    if xmin > xmax:
                        xmin, xmax = xmax, xmin
                    bbox_draw.rectangle([xmin, ymin, xmax, ymax], outline=color, width=4)

        return NemoparseData(text=txt, bbox_json=bbox_data, images=images, tables=tables, bbox_image=base_image) 

    def get_pages(self, filepath):
        file_pages = []
        if isinstance(filepath, io.BytesIO):
            images = convert_bytes_to_images(filepath.read())
            for image in images:
                img_byte_arr = io.BytesIO()
                image.save(img_byte_arr, format='PNG')
                img_byte_arr = img_byte_arr.getvalue()
                file_pages.append(img_byte_arr)
        else:
            if "png" in filepath or "tif" in filepath or "TIF" in filepath:
                with open(filepath, "rb") as image_file:
                    file_pages.append(image_file.read())
            elif "pdf" in filepath:
                images = convert_pdf_to_images(filepath)
                for image in images:
                    img_byte_arr = io.BytesIO()
                    image.save(img_byte_arr, format='PNG')
                    img_byte_arr = img_byte_arr.getvalue()
                    file_pages.append(img_byte_arr)
            else:
                try:
                    images = convert_pdf_to_images(filepath)
                    for image in images:
                        img_byte_arr = io.BytesIO()
                        image.save(img_byte_arr, format='PNG')
                        img_byte_arr = img_byte_arr.getvalue()
                        file_pages.append(img_byte_arr)
                except:
                    raise Exception(f"Unsupported filetype! {filepath}")
        return file_pages

    def process_batch_documents(self, filepaths, use_checkpointing=True, draw_bboxes=True, output_dir="./"):
        file_outputs = []
        for filepath in filepaths:
            output = self.process_document(filepath, draw_bboxes=draw_bboxes)
            if use_checkpointing:
                basename = os.path.basename(filepath)
                print(basename)
                output.save_output(output_dir, basename)
            else:
                file_outputs.append(output)

        return file_outputs

    def process_page(self, page):
        return self._process_image(page)

    def process_document(self, filepath, draw_bboxes=True):
        # Basic check of file type
        file_pages = self.get_pages(filepath) 

        output = NemoparseOutput()
        for page_image in file_pages:
            output.add_output(self._process_image(page_image, draw_bboxes=draw_bboxes))
        return output



import re
import io
import PIL
from typing import List, Tuple

from tqdm import tqdm

try:
    from surya.texify import TexifyPredictor # May need to find an alternative
    USE_OCR = True
except:
    USE_OCR = False
    print("Surya ocr not installed; ocr not available for pptx processing")

from pptx import Presentation
from pptx.shapes.group import GroupShape

from .processor import Processor
from ..output.pptx_output import PptxOutput


try:
    class MarkdownTexifyPredictor(TexifyPredictor):
        def fix_fences(self, text: str) -> str:
            text = re.sub(r'<math display="block">(.*?)</math>',r'$$\1$$', text, flags=re.DOTALL)
            text = re.sub(r'<math>(.*?)</math>',r'$\1$', text, flags=re.DOTALL)
            if re.search(r'<math display="block">', text):
                text = ""
            if re.search(r'<math>', text):
                text = ""
            return text
except:
    class MarkdownTexifyPredictor:
        pass


class PptxProcessor(Processor):
    
    def __init__(self):
        super().__init__()
        if USE_OCR:
            self.latex_predictor = MarkdownTexifyPredictor()
        else:
            self.latex_predictor = None

    def ocr_image(self, image):
        if USE_OCR:
            ocr_output = self.latex_predictor([image])[0]
            if ocr_output.text is not None:
                return ocr_output.text
        return ""

    def process_image(self, image):
        if 'wmf' in image.content_type:
            print("pptx contains a wmf image, skipping image")
            return None
        else:
            byte_io = io.BytesIO(image.blob)
            image_object = PIL.Image.open(byte_io)
            return image_object

    def process_document(self, filepath):
        prs = Presentation(filepath)
        images = []
        slide_texts = []
        for slide in prs.slides:
            slide_text = []
            images.append([])
            for shape in slide.shapes:
                if shape.has_text_frame:
                    for par in shape.text_frame.paragraphs:
                        slide_text.append(par.text)
                if "PIC" in str(shape.shape_type):
                    image = self.process_image(shape.image)
                    if image is not None:
                        image_ocr = self.ocr_image(image)
                        slide_text.append(image_ocr)
                        images[-1].append(image)
                if "GROUP" in str(shape.shape_type):
                    for sub_shape in shape.shapes:
                        if "PIC" in str(sub_shape.shape_type):
                            image = self.process_image(sub_shape.image)
                            if image is not None:
                                image_ocr = self.ocr_image(image)
                                slide_text.append(image_ocr)
                                images[-1].append(image)

                        if sub_shape.has_text_frame:
                            for par in sub_shape.text_frame.paragraphs:
                                slide_text.append(par.text)
            slide_texts.append("\n".join(slide_text))

        metadata = {}
        return PptxOutput(slide_texts, images, metadata)

    def process_batch_documents(self, filepaths):
        pass



import re

def convert_latex_table_to_csv(latex_string):
    cleaned_table = latex_string
    # Remove the LaTeX table environment and any extra formatting
    #cleaned_table = re.sub(r'\\begin\{tabular\}.*?\\end\{tabular\}', '', latex_string, flags=re.DOTALL)
    cleaned_table = re.sub(r'\\begin\{tabular\}\{[clr ]+\}', '', cleaned_table, flags=re.DOTALL)
    cleaned_table = re.sub(r'\\end\{tabular\}', '', cleaned_table, flags=re.DOTALL)
    cleaned_table = re.sub(r'\\[a-zA-Z]+\{.*?\}', '', cleaned_table)  # Remove LaTeX commands
    cleaned_table = re.sub(r'\{[clr ]+\}', '', cleaned_table)  # Remove LaTeX commands
    cleaned_table = re.sub(r'\\hline', '', cleaned_table)  # Remove horizontal lines
    cleaned_table = re.sub(r'\s+', ' ', cleaned_table)  # Normalize whitespace
    cleaned_table = cleaned_table.strip()

    # Split the table into rows
    rows = cleaned_table.split('\\\\')

    # Create csv friendly datastructure
    cleaned_rows = []
    for row in rows:
        # Split each row into columns based on the '&' character
        columns = row.split('&')
        # Strip whitespace from each column
        columns = [col.strip() for col in columns]
        # Write the row to the CSV file
        cleaned_rows.append(columns)

    return cleaned_rows



try:
    from .marker_processor import MarkerProcessor 
except Exception as e:
    print(f"Marker not installed, cannot use MarkerProcessor, {e}")

try:
    from .papermage_processor import PaperMageProcessor
except:
    print("papermage not installed, cannot use PaperMageProcessor")

from .nemoparse_processor import NemoparseProcessor 
from .pptx_processor import PptxProcessor 



import pathlib
import json

from .output import ModelOutput

class PaperMageOutput(ModelOutput):
    def __init__(self, output_data):
        super().__init__()
        self.output_data = output_data
        print(self.output_data)

    def save_output(self, output_directory, filename_base, mode, options=None):
        save_path = pathlib.Path(output_directory)

        if mode == 'bound_single':
            for idx, plotted in enumerate(self.output_data):
                plotted.save(save_path / f'{filename_base}_page_{idx+1}.png', is_overwrite=True, format='PNG')

        elif mode == 'bound_batch':
            for file, single_output in self.output_data.items():
                file_path = pathlib.Path(save_path / file)
                file_path.mkdir(parents=True, exist_ok=True)
                for idx, plotted in enumerate(single_output):
                    plotted.save(file_path / f'{filename_base}_page_{idx+1}.png', is_overwrite=True, format='PNG')

        elif mode == 'extract_single':
            with open(save_path / f"{filename_base}.json", "w+") as f_out:
                print(f"saving to {save_path}")
                json.dump(self.output_data.to_json(), f_out, indent=2)

            for option in options:
                with open(save_path / f"{filename_base}_{option}.txt", "w+") as f:
                    for elem in self.output_data.get_layer(option):
                        f.write(''.join((str(elem),'\n')))
                    f.close()

        elif mode == 'extract_batch':
            for filename, extracted in self.output_data.items():
                file_path = pathlib.Path(save_path / filename)
                file_path.mkdir(parents=True, exist_ok=True)

                with open(file_path / f"{filename_base}.json", "w+") as f_out:
                    json.dump(extracted.to_json(), f_out, indent=2)

                for option in options:
                    with open(file_path / f"{option}.txt", "w+") as f:
                        for elem in extracted.get_layer(option):
                            f.write(''.join((str(elem),'\n')))
                        f.close()




import pdf2image

def convert_pdf_to_images(file_path, dpi=200):
    """
    Converts pdf to images
    Args:
        file_path (str): The path to the PDF file.
    Return:
        list of PIL images
    """
    return pdf2image.convert_from_path(file_path, dpi=dpi)

def convert_bytes_to_images(byte_stream, dpi=200):
    """
    Converts pdf byte stream to images
    Args:
        byte_stream (BytesIO): The byte-stream from the pdf
    Return:
        list of PIL images
    """
    return pdf2image.convert_from_bytes(byte_stream, dpi=dpi)



import pandas as pd

from marker.models import create_model_dict
from marker.settings import settings
from marker.converters.pdf import PdfConverter
from marker.schema.document import Document
from marker.renderers.markdown import MarkdownRenderer
from marker.renderers.markdown import Markdownify
from marker.renderers.markdown import get_formatted_table_text, cleanup_text

from typing import Annotated, Any, Dict, List, Optional, Type, Tuple
from collections import defaultdict
from bs4 import NavigableString
from pydantic import BaseModel

from .processor import Processor
from ..output.marker_output import MarkerOutput


class CustomMarkdownify(Markdownify):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.tables = []

    def convert_table(self, el, text, parent_tags):
        total_rows = len(el.find_all('tr'))
        colspans = []
        rowspan_cols = defaultdict(int)
        for i, row in enumerate(el.find_all('tr')):
            row_cols = rowspan_cols[i]
            for cell in row.find_all(['td', 'th']):
                colspan = int(cell.get('colspan', 1))
                row_cols += colspan
                for r in range(int(cell.get('rowspan', 1)) - 1):
                    rowspan_cols[i + r] += colspan # Add the colspan to the next rows, so they get the correct number of columns
            colspans.append(row_cols)
        total_cols = max(colspans) if colspans else 0

        grid = [[None for _ in range(total_cols)] for _ in range(total_rows)]

        for row_idx, tr in enumerate(el.find_all('tr')):
            col_idx = 0
            for cell in tr.find_all(['td', 'th']):
                # Skip filled positions
                while col_idx < total_cols and grid[row_idx][col_idx] is not None:
                    col_idx += 1

                # Fill in grid
                value = get_formatted_table_text(cell).replace("\n", " ").replace("|", " ").strip()
                rowspan = int(cell.get('rowspan', 1))
                colspan = int(cell.get('colspan', 1))

                if col_idx >= total_cols:
                    # Skip this cell if we're out of bounds
                    continue

                for r in range(rowspan):
                    for c in range(colspan):
                        try:
                            if r == 0 and c == 0:
                                grid[row_idx][col_idx] = value
                            else:
                                grid[row_idx + r][col_idx + c] = '' # Empty cell due to rowspan/colspan
                        except IndexError:
                            # Sometimes the colspan/rowspan predictions can overflow
                            print(f"Overflow in columns: {col_idx + c} >= {total_cols} or rows: {row_idx + r} >= {total_rows}")
                            continue

                col_idx += colspan

        self.tables.append(pd.DataFrame.from_dict(grid))

        markdown_lines = []
        col_widths = [0] * total_cols
        for row in grid:
            for col_idx, cell in enumerate(row):
                if cell is not None:
                    col_widths[col_idx] = max(col_widths[col_idx], len(str(cell)))

        add_header_line = lambda: markdown_lines.append('|' + '|'.join('-' * (width + 2) for width in col_widths) + '|')

        # Generate markdown rows
        added_header = False
        for i, row in enumerate(grid):
            is_empty_line = all(not cell for cell in row)
            if is_empty_line and not added_header:
                # Skip leading blank lines
                continue

            line = []
            for col_idx, cell in enumerate(row):
                if cell is None:
                    cell = ''
                padding = col_widths[col_idx] - len(str(cell))
                line.append(f" {cell}{' ' * padding} ")
            markdown_lines.append('|' + '|'.join(line) + '|')

            if not added_header:
                # Skip empty lines when adding the header row
                add_header_line()
                added_header = True

        # Handle one row tables
        if total_rows == 1:
            add_header_line()

        table_md = '\n'.join(markdown_lines)
        return "\n\n" + table_md + "\n\n"


class CustomMarkdownOutput(BaseModel):
    markdown: str
    images: dict
    metadata: dict
    tables: list


class CustomMarkdownRenderer(MarkdownRenderer):
    page_separator: Annotated[str, "The separator to use between pages.", "Default is '-' * 48."] = "-" * 48
    inline_math_delimiters: Annotated[Tuple[str], "The delimiters to use for inline math."] = ("$", "$")
    block_math_delimiters: Annotated[Tuple[str], "The delimiters to use for block math."] = ("$$", "$$")

    #paginate_output, page_separator, inline_math_delimiters, block_math_delimiters
    @property
    def md_cls(self):
        return CustomMarkdownify(
            paginate_output=self.paginate_output,
            page_separator=self.page_separator,
            heading_style="ATX",
            bullets="-",
            escape_misc=False,
            escape_underscores=True,
            escape_asterisks=True,
            escape_dollars=True,
            sub_symbol="<sub>",
            sup_symbol="<sup>",
            inline_math_delimiters=self.inline_math_delimiters,
            block_math_delimiters=self.block_math_delimiters,
            html_tables_in_markdown=False
        )

    def __call__(self, document: Document) -> CustomMarkdownOutput:
        document_output = document.render()
        full_html, images = self.extract_html(document, document_output)
        _md_cls = self.md_cls
        markdown = _md_cls.convert(full_html)
        markdown = cleanup_text(markdown)
        tables = _md_cls.tables
        return CustomMarkdownOutput(
            markdown=markdown,
            images=images,
            metadata=self.generate_document_metadata(document, document_output),
            tables=tables
        )


class CustomPdfConverter(PdfConverter):
    """
    A converter for processing and rendering PDF files into Markdown, JSON, HTML and other formats.
    """
    def __init__(
        self,
        artifact_dict: Dict[str, Any],
        renderer = None, 
        processor_list: Optional[List[str]] = None,
        llm_service: str | None = None,
        config=None,
    ):
        super().__init__(
                        artifact_dict=artifact_dict,
                        processor_list=processor_list,
                        renderer=renderer if isinstance(renderer, str) else None,
                        llm_service=llm_service,
                        config=config)

        if (not isinstance(renderer, str)) and (renderer is not None):
            self.renderer = renderer


class MarkerProcessor(Processor):
    
    def __init__(self):
        super().__init__()
        models = create_model_dict()
        config = {}
        config["format_lines"] = True
        config["force_ocr"] = True
        self.converter = CustomPdfConverter(
                        artifact_dict = models,
                        renderer=CustomMarkdownRenderer,
                        config=config,
                        )

    def process_document(self, filepath):
        output = self.converter(filepath)
        return MarkerOutput(output)

    def process_batch_documents(self, filepaths):
        file_outputs = []
        for filepath in filepaths:
            output = self.process_document(filepath)
            file_outputs.append(output)
            #try:
            #    output = self.process_document(filepath)
            #    file_outputs.append(output)
            #except:
            #    print(f"Failed to parse document: {filepath}")

        return file_outputs



import os
import json

from marker.settings import settings

from .output import ModelOutput


class MarkerOutput(ModelOutput):
    def __init__(self, output_data):
        super().__init__()
        self.text = output_data.markdown.encode(settings.OUTPUT_ENCODING, errors="replace").decode(
            settings.OUTPUT_ENCODING
        )
        self.images: list = output_data.images
        self.tables: list = output_data.tables
        self.metadata = output_data.metadata

    def save_output(self, output_directory, filename_base):
        with open(
            os.path.join(output_directory, f"{filename_base}.md"),
            "w+",
            encoding=settings.OUTPUT_ENCODING,
        ) as f:
            f.write(self.text)
        with open(
            os.path.join(output_directory, f"{filename_base}_meta.json"),
            "w+",
            encoding=settings.OUTPUT_ENCODING,
        ) as f:
            f.write(json.dumps(self.metadata, indent=2))

        for img_name, img in self.images.items():
            img.save(os.path.join(output_directory, f"{filename_base}_{img_name}"), settings.OUTPUT_IMAGE_FORMAT)

        for i, table in enumerate(self.tables):
            table_name = f"{filename_base}_table_{i}.csv"
            table.to_csv(os.path.join(output_directory, table_name))



from abc import ABC, abstractmethod

class Processor(ABC):
    def __init__(self):
        pass

    @abstractmethod
    def process_document(self, filepath):
        pass

    @abstractmethod
    def process_batch_documents(self, filepaths):
        pass
